from __future__ import annotations
from math import exp, log10, log, pow
from typing import Callable


def db_to_volume_value(db: float):
    '''
    Maps a dB value to a volume value (0 - 1).
    '''
    return exp((db - 6) * (1 / 20)) if db > -100 else 0


def gain_to_db(gain: float):
    '''
    @param gain A value between 0 - 2, corresponding to -inf to +6dB, gain == 1 equals dB == 0.0
    '''
    if (gain <= 0):
        return -100

    return 20 * log10(gain)


def volume_value_to_db(volume: float):
    '''
    @param volume A value between 0 - 1, corresponding to -inf to +6 dB.
    '''
    return 20 * log(volume) + 6 if volume > 0 else -100.0


def volume_value_to_gain(volume: float):
    '''
    @param volume A value between 0 - 1, corresponding to -inf to +6 dB.
    '''
    return pow(10, (20 * log(volume) + 6) * (1 / 20)) if volume > 0 else 0


def gain_to_volume_value(gain: float):
    '''
    @param gain A value between 0 - 2, corresponding to -inf to +6dB, gain == 1 equals dB == 0.0
    '''
    return exp((20 * log10(gain) - 6) * (1 / 20)) if gain > 0 else 0


def greater_equal(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns the index of the first item in the array >= val. This is a successor query which also returns the item if present.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = high + 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p >= 0:
            i = m
            high = m - 1
        else:
            low = m + 1
    return i


def greater_than(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = high + 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p > 0:
            i = m
            high = m - 1
        else:
            low = m + 1
    return i


def lower_than(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = low - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p < 0:
            i = m
            low = m + 1
        else:
            high = m - 1
    return i


def lower_equal(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = low - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if (p <= 0):
            i = m
            low = m + 1
        else:
            high = m - 1
    return i


def eq(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p == 0:
            return m
        if p <= 0:
            low = m + 1
        else:
            high = m - 1
    return -1



from tuneflow_py.descriptors.text import LabelText
from tuneflow_py.descriptors.common import PluginInfo
from tuneflow_py.descriptors.param import ParamDescriptor
from tuneflow_py.descriptors.audio_plugin import AudioPluginDescriptor
from typing import Optional, Any, List, Dict
from tuneflow_py.models.song import Song


class ReadAPIs:
    '''
    Read-only APIs used in `init` and `run` methods of a plugin.
    '''

    def get_system_locale(self):
        raise Exception("Not implemented")

    def translate_label(self, label_text: LabelText):
        raise Exception("Not implemented")

    def serialize_song(self, song: Song):
        raise Exception("Not implemented")

    def deserialize_song(self, encoded_song: str):
        raise Exception("Not implemented")

    def get_available_audio_plugins(self) -> List[AudioPluginDescriptor]:
        raise Exception("Not implemented")


class TuneflowPlugin:
    '''
    The base class of a plugin.

    All plugins should be a sub-class of this plugin in order to run in the pipeline.
    '''

    def __init__(self) -> None:
        self.params_result_internal = {}

    @staticmethod
    def provider_id() -> str:
        """
        The unique id to identify the plugin provider.

        For example:
        `friday-core`, `some-corp`, etc.
        """
        raise Exception("provider_id must be overwritten.")

    @staticmethod
    def plugin_id() -> str:
        '''
        The unique plugin id under the provider's namespace.

        For example:
        `melody-extractor`, `tune-analyzer`, etc.
        '''
        raise Exception("plugin_id must be overwritten.")

    @staticmethod
    def provider_display_name() -> LabelText:
        '''
        The display name of the provider.
        '''
        raise Exception("provider_display_name must be overwritten.")

    @staticmethod
    def plugin_display_name() -> LabelText:
        '''
        The display name of the plugin.
        '''
        raise Exception("plugin_display_name must be overwritten.")

    @staticmethod
    def plugin_description() -> Optional[LabelText]:
        '''
        The description of this plugin.
        '''
        return None

    @staticmethod
    def plugin_info() -> Optional[PluginInfo]:
        return None

    @staticmethod
    def allow_reset():
        '''
        Whether to allow users to reset all parameters of this plugin.
        '''
        return False

    def init(self, song: Song, read_apis: ReadAPIs):
        '''
        Initializes the plugin instance.

        Override this method to initialize your plugin before it starts running.
        '''
        pass

    def params(self) -> Dict[str, ParamDescriptor]:
        '''
        Specify params to get from user input.

        Param input widgets will be displayed on the UI, and the inputs will be collected and fed into @run method.

        If you don't need any param, return `{}`;
        '''
        return {}

    def allow_manual_apply_adjust(self):
        '''
        Whether the user can manually apply this plugin and go back to adjust it.
        Enable this when you want the user to frequently toggle this plugin on and off
        to see the difference.

        For example: A plugin that divides a track into two, you want the user to
        easily switch between the plugin is on or off to see what's going on.
        '''
        return False

    def run(self, song: Song, params: Dict[str, Any], read_apis: ReadAPIs):
        '''
        The main logic here.

        Args:
            `song`: The song that is being processed. You can directly modify the song
                by calling its methods.
            `params`: The results collected from user input specified by the `params` method.
        '''
        pass

    # =====================================
    # NO OVERWRITE BELOW
    # =====================================

    @classmethod
    def create(cls, song: Song, read_apis: ReadAPIs):
        plugin = cls()
        plugin.reset_internal()
        plugin.init(song, read_apis=read_apis)
        return plugin

    def reset_internal(self):
        params = self.params()
        for key in params.keys():
            param_descriptor = params[key]
            self.params_result_internal[key] = param_descriptor["defaultValue"]



from tuneflow_py.models.song import Song
from miditoolkit.midi import MidiFile
from pathlib import PurePath, Path
from io import BytesIO
import unittest


class TestSong(unittest.TestCase):
    def test_import_midi(self):
        midi_obj = MidiFile(filename=PurePath(
            Path(__file__).parent, Path('caravan.test.golden.mid')))
        song = Song.from_midi(midi_obj=midi_obj)
        # Test tempos
        self.assertEqual(song.get_tempo_event_count(), 8)
        self.assertAlmostEqual(song.get_duration(), 595.0945734687816)
        # Test time signatures
        self.assertEqual(song.get_time_signature_event_count(), 3)
        self.assertEqual(
            song.get_time_signature_event_at(1).get_numerator(), 4)
        self.assertEqual(song.get_time_signature_event_at(
            1).get_denominator(), 2)
        # Test tracks
        first_track = song.get_track_at(0)
        self.assertEqual(first_track.get_clip_count(), 1)
        # Test non-piano track
        # Test track volume, pan
        self.assertAlmostEqual(
            first_track.get_volume_in_db(), -0.6665511534374842)
        self.assertAlmostEqual(song.get_track_at(7).get_pan(), 63)
        # Test track instrument
        self.assertEqual(first_track.get_instrument().program, 65)
        self.assertEqual(first_track.get_instrument().is_drum, False)
        # Test drum track
        last_track = song.get_track_at(song.get_track_count()-1)
        self.assertEqual(last_track.get_instrument().program, 0)
        self.assertEqual(last_track.get_instrument().is_drum, True)
        # TODO: Test track automation
        first_clip = first_track.get_clip_at(0)
        self.assertEqual(first_clip.get_note_count(), 673)
        self.assertEqual(first_clip.get_clip_start_tick(), 107520)
        self.assertEqual(first_clip.get_clip_end_tick(), 1326719)
        first_note = first_clip.get_raw_note_at(0)
        self.assertEqual(first_note.get_pitch(), 52)
        self.assertEqual(first_note.get_start_tick(), 107520)
        self.assertEqual(first_note.get_end_tick(), 108159)
        self.assertEqual(first_note.get_velocity(), 115)
        self.assertEqual(first_note.get_start_time(), 47.659584045410156)
        self.assertAlmostEqual(first_note.get_end_time(), 47.94282913208008)
        self.assertEqual(song.last_tick, 1327199)
        self.assertAlmostEqual(song.duration, 595.0945734687816)

    def test_export_midi(self):
        golden_midi_path = PurePath(
            Path(__file__).parent, Path('caravan.test.golden.mid'))
        export_midi_path = PurePath(
            Path(__file__).parent, Path('caravan.test.test.mid'))
        midi_obj = MidiFile(filename=golden_midi_path)
        song = Song.from_midi(midi_obj=midi_obj)
        export_file = open(export_midi_path, 'wb')
        exported_midi = song.to_midi()
        exported_midi.dump(file=export_file)
        export_file.close()

        self.assertEqual(len(midi_obj.tempo_changes),
                         len(exported_midi.tempo_changes))
        for i in range(len(midi_obj.tempo_changes)):
            expected_tempo = midi_obj.tempo_changes[i]
            actual_tempo = exported_midi.tempo_changes[i]
            self.assertEqual(expected_tempo.time, actual_tempo.time)
            self.assertAlmostEqual(expected_tempo.tempo,
                                   actual_tempo.tempo, places=4)
        self.assertEqual(len(midi_obj.time_signature_changes),
                         len(exported_midi.time_signature_changes))
        for i in range(len(midi_obj.time_signature_changes)):
            expected_time_signature = midi_obj.time_signature_changes[i]
            actual_time_signature = exported_midi.time_signature_changes[i]
            self.assertEqual(expected_time_signature.time,
                             actual_time_signature.time)
            self.assertEqual(expected_time_signature.numerator,
                             actual_time_signature.numerator)
            self.assertEqual(expected_time_signature.denominator,
                             actual_time_signature.denominator)

        self.assertEqual(len(midi_obj.instruments),
                         len(exported_midi.instruments))
        for i in range(len(midi_obj.instruments)):
            expected_track = midi_obj.instruments[i]
            actual_track = exported_midi.instruments[i]
            self.assertEqual(expected_track.program, actual_track.program)
            self.assertEqual(expected_track.is_drum, actual_track.is_drum)
            self.assertEqual(len(expected_track.notes),
                             len(actual_track.notes))
            for j in range(len(expected_track.notes)):
                expected_note = expected_track.notes[j]
                actual_note = actual_track.notes[j]
                self.assertEqual(expected_note.start, actual_note.start)
                self.assertEqual(expected_note.end, actual_note.end)
                self.assertEqual(expected_note.velocity, actual_note.velocity)
                self.assertEqual(expected_note.pitch, actual_note.pitch)
            # TODO: Test track automation.


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2


class TempoEvent:
    def __init__(self, ticks: int | None = None, bpm: float | None = None, time: float | None = None, proto: song_pb2.TempoEvent | None = None):
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.TempoEvent()
            self._proto.ticks = ticks
            self._proto.bpm = bpm
            if time is not None:
                self._proto.time = time

    def get_ticks(self) -> int:
        return self._proto.ticks

    def get_bpm(self) -> float:
        return self._proto.bpm

    def get_time(self) -> float:
        return self._proto.time



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.note import Note
from nanoid import generate as generate_nanoid

ClipType = song_pb2.ClipType


class Clip:
    def __init__(self, proto: song_pb2.Clip | None = None) -> None:
        self._proto = proto if proto is not None else song_pb2.Clip()

    def get_note_count(self):
        return len(self._proto.notes)

    def get_raw_notes(self):
        for note_proto in self._proto.notes:
            yield Note(proto=note_proto)

    def get_raw_note_at(self, index: int):
        return Note(proto=self._proto.notes[index])

    def get_type(self) -> int:
        return self._proto.type

    def delete_note_at(self, index: int):
        self._proto.notes.pop(index)

    def get_clip_start_tick(self) -> int:
        return self._proto.clip_start_tick

    def get_clip_end_tick(self) -> int:
        return self._proto.clip_end_tick

    @staticmethod
    def is_note_in_clip(
        note_start_tick: int,
        note_end_tick: int,
        clip_start_tick: int,
        clip_end_tick: int,
    ):
        '''
        Note is in the clip when:
        * The note starts at or after the clip start.
        * The note has overlapping part with the clip.
        * The overlapping part is greater than 1 tick, i.e. the note does not start from the end of the clip or end at the start of the clip.
        '''
        # If the clip starts at 0, notes that start before 0 will start at 0.
        return (
            (note_start_tick >= clip_start_tick or (clip_start_tick == 0 and note_start_tick <= 0)) and
            note_start_tick < clip_end_tick and
            note_end_tick > note_start_tick
        )

    @staticmethod
    def _generate_clip_id():
        return generate_nanoid(size=10)



from tuneflow_py.descriptors.text import LabelText
from tuneflow_py.descriptors.widget import WidgetDescriptor
from typing import Any, Optional
from typing_extensions import TypedDict
from enum import Enum


class InjectSource(Enum):
    SelectedTrackIds = 1
    ''' A string that represents the ids of the current selected tracks. '''

    SelectedClipInfos = 2
    ''' A `list[ClipInfo]` that represents the `ClipInfo`s of the current selected clips. '''

    TickAtPlayhead = 3
    ''' A number that represents the current tick that the playhead is at. '''

    EditingClipInfo = 4
    ''' The `ClipInfo` of the clip that is being edited in the MIDI editor. '''

    EditingNoteIds = 5
    ''' The ids of the notes that is being edited in the MIDI editor. '''

    TickAtPlayheadSnappedToBeat = 6
    ''' A number that represents the start tick of the beat where the playhead is at. '''


class ParamDescriptor(TypedDict, total=False):
    '''
    Definition of the config for a single param.

    Note that the members here use camel case to be
    consistent with the typescript plugins.
    '''

    displayName: LabelText
    ''' The name to display on the UI. '''

    widget: WidgetDescriptor
    ''' Configuration of the widget to display on the UI. '''

    defaultValue: Any
    ''' The default value of the param. '''

    adjustable: Optional[bool]
    '''
    Whether this param is adjustable. If the param is not adjustable, the controllable
    part is not shown.

    Defaults to true.
    '''

    hidden: Optional[bool]
    ''' Whether this param is hidden completely. '''

    optional: Optional[bool]
    ''' Whether this param can be left undefined or null. '''

    description: Optional[LabelText]
    ''' Explaining what this parameter is for. '''

    injectFrom: InjectSource
    '''
    Injects the value from the editor at the time the plugin runs.
    
    If specified, the editor will inject the value specified by the `InjectSource`,
    if you want to overwrite the value provided by the editor,
    you can use initPluginInstance.
    '''

    adjustableWhenPluginIsApplied: Optional[bool]
    '''
    If set to true, this param can still be adjusted
    when the plugin is applied. However, changing the param
    will invalidate all the changes after this plugin.
    '''



from typing_extensions import TypedDict

class AudioPluginDescriptor(TypedDict, total=False):
    '''
    Descriptor of an audio plugin.
    '''
    
    category: str
    '''
    Category of the plugin.
    e.g.: 'Instrument|Synth', 'Fx|EQ', etc.
    '''

    descriptiveName: str
    ''' A more detailed name. '''

    id: str
    ''' An internal id of the plugin. '''

    isInstrument: str
    '''
    Whether this plugin can be used as an instrument to generate
    sound from MIDI notes.
    '''

    manufacturerName: str
    ''' Manufacturer of the plugin. '''

    name: str
    '''
    A short name of the plugin.
    Use this when creating `AudioPlugin` instances.
    '''

    pluginFormatName: str
    ''' Format of the plugin. e.g. 'VST3', 'AudioUnit', etc. '''

    version: str
    ''' Version of the plugin. '''
